import os

import pandas as pd
from os.path import basename, isfile, splitext

from raven_preprocess.data_source_info import DataSourceInfo, SOURCE_TYPE_FILE
from raven_preprocess.msg_util import msg, msgt
from raven_preprocess.file_format_constants import \
    (CSV_FILE_EXT, TAB_FILE_EXT,
     XLS_FILE_EXT, XLSX_FILE_EXT,
     ACCEPTABLE_EXT_LIST,
     get_mime_type)


class FileFormatUtil(object):
    def __init__(self, input_file, **kwargs):
        """Identify file type: csv, tab, etc"""

        self.input_file = input_file

        self.job_id = kwargs.get('job_id')

        self.dataframe = None
        self.data_source_info = None

        self.file_basename = None
        self.fname_ext_check = None

        #
        self.has_error = False
        self.error_message = None

        self.check_basic_file_info()
        self.load_dataframe()

    def add_error(self, err):
        self.has_error = True
        self.error_message = err

    def check_basic_file_info(self):
        """split the filename"""
        if self.has_error:
            return

        # Is the file specified?
        #
        if not self.input_file:
            self.add_error('The "input_file" was %s' % self.input_file)
            return

        # Does the file exist?
        #
        if isinstance(self.input_file, str):
            # "self.input_file" is file path
            if not isfile(self.input_file):
                self.add_error('The file was not found: [%s]' % self.input_file)
                return

            self.filesize = os.stat(self.input_file).st_size
            # print('filesize ', self.filesize)
            self.file_basename = basename(self.input_file)
            # print('file basename ', self.file_basename)

        else:
            # "self.input_file" is Django FileField
            self.filesize = self.input_file.size
            # print('filesize ', self.filesize)
            self.file_basename = basename(self.input_file.name)
            # print('file basename ', self.file_basename)

        if self.filesize == 0:
            self.add_error('The file size is zero: [%s]' % self.input_file)
            return

        # file basename with extension
        #

        # file extension
        #
        _fname_base, fname_ext = splitext(self.file_basename)

        if fname_ext:
            self.fname_ext_check = fname_ext.lower()
        else:
            self.add_error(\
                ('The "input_file" did not have an extension: %s') % \
                self.input_file)
            return


        # Is this a known/accepted extension
        #
        if self.fname_ext_check not in ACCEPTABLE_EXT_LIST:
            self.add_error(self.get_unaccepted_file_err())
            return

        # Set the mimetype based on the extension
        #
        source_format = get_mime_type(self.fname_ext_check)

        # Create the informational DataSourceInfo object
        #
        self.data_source_info = DataSourceInfo(name=self.file_basename,
                                               source_type=SOURCE_TYPE_FILE,
                                               source_format=source_format,
                                               filesize=self.filesize)


    def load_dataframe(self):
        """Load the file to a dataframe"""
        if self.has_error:
            return

        if self.fname_ext_check == TAB_FILE_EXT:
            self.set_dataframe('\t')

        elif self.fname_ext_check == CSV_FILE_EXT:
            self.set_dataframe()

        elif self.fname_ext_check == XLS_FILE_EXT:
            self.set_dataframe(is_excel=True)

        elif self.fname_ext_check == XLSX_FILE_EXT:
            self.set_dataframe(is_excel=True)

        else:
            self.add_error(self.get_unaccepted_file_err())


    def get_unaccepted_file_err(self):
        """Error message used multiple times"""
        name_for_err = self.input_file
        if self.file_basename:
            name_for_err = self.file_basename

        return ('We currently do not process this file type.'
                ' Please use a file with one of the following'
                ' extensions: %s'
                '\nFile name: %s') % \
                (ACCEPTABLE_EXT_LIST, name_for_err)


    def set_dataframe(self, delimiter=None, is_excel=False):
        """Retrieve a pandas dataframe based on format"""
        try:
            if is_excel:
                self.dataframe = pd.read_excel(self.input_file)
            else:
                self.dataframe = pd.read_csv(self.input_file,
                                             delimiter=delimiter)

        except pd.errors.ParserError as err_obj:
            err_msg = ('Failed to load csv file (pandas ParserError).'
                       '\n- File: %s'
                       '\n- %s') % \
                      (self.input_file, err_obj)
            self.add_error(err_msg)

        except PermissionError as err_obj:
            err_msg = ('No read prermission on this file:'
                       '\n- File: %s'
                       '\n- %s') % \
                      (self.input_file, err_obj)
            self.add_error(err_msg)

        except Exception as err_obj:
            err_msg = ('Failed to load file.'
                       '\n- File: %s'
                       '\n- %s') % \
                      (self.input_file, err_obj)
            self.add_error(err_msg)



"""Module for a variable's display settings"""
from collections import OrderedDict
import json

import raven_preprocess.col_info_constants as col_const
import raven_preprocess.update_constants as update_const
from raven_preprocess.column_info import ColumnInfo
from raven_preprocess.np_json_encoder import NumpyJSONEncoder
from raven_preprocess.version_number_util import VersionNumberUtil

ALL_VARIABLE_ATTRIBUTES = [x[0] for x in ColumnInfo.get_variable_labels()]
EDITABLE_ATTRIBUTES = ColumnInfo.get_editable_column_labels()

class VariableDisplayUtil(object):

    def __init__(self, preprocess_json, update_json):
        """Init with a pandas dataframe"""
        assert isinstance(preprocess_json, dict),\
            "preprocess_json must be a dict/OrderedDict"
        assert isinstance(update_json, dict),\
            "update_json must be a dict/OrderedDict"

        self.preprocess_json = preprocess_json
        self.update_json = update_json

        # count the updates
        self.update_display_cnt = 0
        self.update_value_cnt = 0

        # for error handling
        self.has_error = False
        self.error_messages = []
        self.original_json = None
        self.access_obj_original = None
        self.access_obj_original_display = None

        # call the display function
        self.update_preprocess_data()

    def add_error_message(self, err_msg):
        """Add error message"""
        print(err_msg)
        self.has_error = True
        self.error_messages.append(err_msg)

    def get_updated_metadata(self, as_string=False):
        """Return the modified metadata--which is in the 'original_json' """
        assert self.has_error is False, \
              "Make sure that 'has_error' is False before using this method"

        if as_string:
            return json.dumps(self.original_json,
                              indent=4,
                              cls=NumpyJSONEncoder)

        return self.original_json

    @staticmethod
    def get_default_settings():
        """Return the initial preprocess settings"""
        return OrderedDict(viewable=True,
                           omit=[],
                           images=[])


    def get_error_messages(self):
        """Return the list of error messages"""
        return self.error_messages


    def run_basic_checks(self):
        """Do some sanity checks--replace this with JSON schema checks..."""
        if col_const.PREPROCESS_ID not in self.update_json:
            self.add_error_message(\
                "A '%s' was not found in the update JSON" % col_const.PREPROCESS_ID)
            return False

        if col_const.SELF_SECTION_KEY not in self.preprocess_json:
            self.add_error_message(\
                "A '%s' was not found in the preprocess JSON" % col_const.SELF_SECTION_KEY)
            return False

        if col_const.PREPROCESS_ID not in self.preprocess_json[col_const.SELF_SECTION_KEY]:
            self.add_error_message(\
                "A '%s.%s' was not found in the preprocess JSON" % \
                (col_const.SELF_SECTION_KEY, col_const.PREPROCESS_ID))
            return False

        #print('-/\\' * 40)
        #print('update: %s' % self.update_json[col_const.PREPROCESS_ID])
        #print('preprocess_json: %s' % #self.preprocess_json[col_const.SELF_SECTION_KEY][col_const.PREPROCESS_ID])
        #print('-/\\' * 40)

        if not self.update_json[col_const.PREPROCESS_ID] == \
            self.preprocess_json[col_const.SELF_SECTION_KEY][col_const.PREPROCESS_ID]:
            self.add_error_message(\
                ('The "{0}" in the update does not match the'
                 ' "{1}.{0}" in the preprocess metadata').format(\
                 col_const.PREPROCESS_ID, col_const.SELF_SECTION_KEY))
            return False

        if col_const.VARIABLES_SECTION_KEY in self.preprocess_json:
            self.access_obj_original = self.preprocess_json[col_const.VARIABLES_SECTION_KEY]
        else:
            self.add_error_message(
                '"%s" section not found in the preprocess data' % col_const.VARIABLES_SECTION_KEY)
            return False

        if col_const.VARIABLE_DISPLAY_SECTION_KEY in self.preprocess_json:
            self.access_obj_original_display = \
                self.preprocess_json[col_const.VARIABLE_DISPLAY_SECTION_KEY]
        else:
            self.add_error_message(\
                '"%s" section not found in the preprocess data' % \
                col_const.VARIABLE_DISPLAY_SECTION_KEY)
            return False


        return True

    def is_major_update(self):
        """updates to the values are "major", updates to the display are "minor" """
        if self.update_value_cnt > 0:
            return True
        return False

    def update_preprocess_data(self):
        """Iterate through the update_json and call omit,viewable,label functions"""
        if not self.run_basic_checks():
            return False, self.get_error_messages

        self.original_json = self.preprocess_json

        if update_const.VARIABLE_UPDATES in self.update_json:
            variable_updates = self.update_json[update_const.VARIABLE_UPDATES]
        else:
            variable_updates = None
            self.add_error_message(
                '%s not found in Update file' % update_const.VARIABLE_UPDATES)
            return False, self.get_error_messages()


        # Iterate through the columns and make updates
        #
        cols_to_update = list(variable_updates)
        for varname in cols_to_update:

            # Check for changes to the omit list
            #
            if update_const.OMIT_KEY in variable_updates[varname]:
                self.update_omit_list(\
                        varname,
                        variable_updates[varname][update_const.OMIT_KEY])

            # Make updates to the viewable section for a variable
            #
            if update_const.VIEWABLE_KEY in variable_updates[varname]:
                self.update_viewable(\
                        varname,
                        variable_updates[varname][update_const.VIEWABLE_KEY])

            # Make value updates
            #
            if update_const.VALUE_UPDATES_KEY in variable_updates[varname]:
                value_update_dict = variable_updates[varname][update_const.VALUE_UPDATES_KEY]
                self.modify_original(varname, value_update_dict)

        # Check if any updates were made...
        #
        if (self.update_value_cnt + self.update_display_cnt) == 0:
            user_msg = ('The update request will not change'
                        ' the current preprocess metadata.'
                        ' A new version was NOT created')
            self.add_error_message(user_msg)

        # Update the version number
        #
        self.update_version_number()

        # Were there errors?
        #
        if self.has_error:
            return False, self.get_error_messages()

        return True, self.get_updated_metadata()


    def update_version_number(self):
        """Bump the version number"""
        if col_const.SELF_SECTION_KEY not in self.original_json:
            user_msg = ('The "%s" section was not found in the'
                        ' preprocess metadata') % \
                        (col_const.SELF_SECTION_KEY)
            self.add_error_message(user_msg)
            return False

        if col_const.VERSION_KEY not in self.original_json[col_const.SELF_SECTION_KEY]:
            user_msg = ('The "%s%s" was not found in the'
                        ' preprocess metadata') % \
                        (col_const.SELF_SECTION_KEY, col_const.VERSION_KEY)
            self.add_error_message(user_msg)
            return False

        success, updated_or_err = VersionNumberUtil.update_version_number(\
                                        self.original_json,
                                        self.is_major_update())

        if not success:
            self.add_error_message(updated_or_err)
            return False

        return True

    def is_in_display_section(self, varname):
        """Make sure the variable is in the 'variable_display' section"""
        assert self.access_obj_original_display is not None,\
                ('The self.access_obj_original_display must'
                 ' be populated before using this method')

        if varname not in self.access_obj_original_display:
            user_msg = ('The variable "%s" was not found in the "%s" section'
                        ' of the metadata.') % \
                        (varname, col_const.VARIABLE_DISPLAY_SECTION_KEY)
            self.add_error_message(user_msg)
            return False

        return True


    def is_in_variables_section(self, varname):
        """Make sure the variable is in the 'variables' section"""
        assert self.access_obj_original is not None,\
                ('The self.access_obj_original must'
                 ' be populated before using this method')

        if varname not in self.access_obj_original:
            user_msg = ('The variable "%s" was not found in the "%s" section'
                        ' of the metadata.') % \
                        (varname, col_const.VARIABLES_SECTION_KEY)
            self.add_error_message(user_msg)
            return False

        return True

    def update_viewable(self, varname, viewable_val):
        """Update the viewable parameter within the variable display section"""
        assert self.access_obj_original_display is not None,\
            ('The self.access_obj_original_display must'
             ' be populated before using this method')
        if not self.is_in_display_section(varname):
            return

        if viewable_val is None:
            return

        # Is it a legimate value?
        if viewable_val not in (True, False):
            user_msg = ('For variable "%s" the value for "%s" must'
                        ' be "true" or "false"'
                        ' (Hint: make sure it is not a string)') % \
                        (varname, update_const.VIEWABLE_KEY)
            self.add_error_message(user_msg)
            return

        # Is it an update?
        #
        if self.access_obj_original_display[varname][update_const.VIEWABLE_KEY] ==\
            viewable_val:
            # Same value, no update to make
            return

        self.update_display_cnt += 1
        self.access_obj_original_display[varname][update_const.VIEWABLE_KEY] = \
            viewable_val


    def update_omit_list(self, varname, omit_list):
        """Update the omit_list in the preprocess metadata.  """
        assert self.access_obj_original_display is not None,\
            ('The self.access_obj_original_display must'
             ' be populated before using this method')
        if not self.is_in_display_section(varname):
            return

        if omit_list is None:
            return

        # Make sure all of the omit variables are valid
        #
        err_found = False
        for omit_var in omit_list:
            if omit_var not in ALL_VARIABLE_ATTRIBUTES:
                err_found = True
                err_msg = ('Variable "%s", which is in the %s list for'
                           ' %s does not exist.') %\
                           (omit_var, update_const.OMIT_KEY, varname)
                self.add_error_message(err_msg)

        if err_found:
            return

        if set(self.access_obj_original_display[varname][update_const.OMIT_KEY]) ==\
                set(omit_list):
            # nothing to change, keep going
            pass
        else:
            self.access_obj_original_display[varname][update_const.OMIT_KEY] = omit_list
            self.update_display_cnt += 1


    def modify_original(self, varname, value_update_dict):
        """Make updates to the original preprocess file for a single variable"""
        assert self.access_obj_original is not None,\
            ('The self.access_obj_original must'
             ' be populated before using this method')
        if not self.is_in_variables_section(varname):
            return

        if not value_update_dict:
            return

        # Iterate through the udpates...
        #
        for update_var, update_value in value_update_dict.items():

            if not self.is_update_value_valid(varname, update_var, update_value):
                continue
            elif self.access_obj_original[varname][update_var] != update_value:
                # Is this a change from the current value? YES!
                #
                self.access_obj_original[varname][update_var] = update_value
                self.update_value_cnt += 1



    def is_update_value_valid(self, varname, update_var, update_value):
        """Check if the value for a new variable is valid"""
        assert update_var, 'The "update_var" cannot be None'

        # Make sure this is an editable variable
        #
        if update_var not in EDITABLE_ATTRIBUTES:
            err_msg = ('For the variable "%s", the value for "%s"'
                       ' is not editable.  Editable variables are:'
                       ' %s') % \
                       (varname, update_var, EDITABLE_ATTRIBUTES)
            self.add_error_message(err_msg)
            return False

        # If appropriate, check for a valid "nature" value
        #
        if update_var == col_const.NATURE_LABEL and \
            not ColumnInfo.is_valid_nature(update_value):

            err_msg = ('For the variable "%s", the value for "%s"'
                       ' is not valid.  Valid values are:'
                       ' %s') % \
                       (varname, col_const.NATURE_LABEL, col_const.NATURE_VALUES)
            self.add_error_message(err_msg)
            return False

        # If appropriate, check for a valid "numchar" value
        #
        if update_var == col_const.NUMCHAR_LABEL and \
            not ColumnInfo.is_valid_numchar(update_value):

            err_msg = ('For the variable "%s", the value for "%s"'
                       ' is not valid.  Valid values are:'
                       ' %s') % \
            (varname, col_const.NUMCHAR_LABEL, col_const.NUMCHAR_VALUES)
            self.add_error_message(err_msg)
            return False

        return True



from os.path import abspath, dirname, join, normpath, isdir, isfile
import sys
import pandas as pd


PREPROCESS_DIR = dirname(dirname(abspath(__file__)))
INPUT_DIR = join(PREPROCESS_DIR, 'input')
# add the 'code' directory to the sys path
sys.path.append(join(PREPROCESS_DIR, 'code'))

from raven_preprocess.msg_util import msg, msgt, dashes
from raven_preprocess.type_guess_util import TypeGuessUtil
import raven_preprocess.col_info_constants as col_const

def main():
    df = pd.DataFrame.from_csv(join(INPUT_DIR, 'test_file_01.csv'),
                               index_col=None)

    print(df.columns)
    #df.columns
    col = df['Id']

    col_oil = df['oil']
    print('isnull', col_oil.isnull())
    print('isnull', col_oil.isnull().sum(), int(col_oil.isnull().sum()))
    print('valid', col_oil.count())
    return
    #val_counts = col.value_counts(sort=False, ascending=False)

    pairs = []
    for col_val, val_cnt in col.value_counts(sort=True).iteritems():
        print(col_val, val_cnt)
        pairs.append((col_val, val_cnt))

    #print(sorted(pairs, key=lambda x: (x[1], x[0]), reverse=True))

if __name__ == '__main__':
    main()



from __future__ import print_function
import json, time, datetime
from collections import OrderedDict
import decimal
import pandas as pd
import os
from os.path import isfile
from raven_preprocess.basic_utils.basic_response import (ok_resp, err_resp)


import raven_preprocess.col_info_constants as col_const
from raven_preprocess.np_json_encoder import NumpyJSONEncoder
from raven_preprocess.type_guess_util import TypeGuessUtil
from raven_preprocess.summary_stats_util import SummaryStatsUtil
from raven_preprocess.column_info import ColumnInfo
from raven_preprocess.plot_values import PlotValuesUtil
from raven_preprocess.variable_display_util import VariableDisplayUtil
from raven_preprocess.dataset_level_info_util import DatasetLevelInfo
from raven_preprocess.file_format_util import FileFormatUtil


KEY_JSONLD_CITATION = 'jsonld_citation'


class PreprocessRunner(object):
    """Preprocess relatively small files using pandas"""

    def __init__(self, dataframe, **kwargs):
        """Init with a pandas dataframe

        optional kwargs:
        job_id - id of a PreprocessJob object
        jsonld_citation - jsonld_citation as an OrderedDict
        """
        self.data_frame = dataframe
        self.job_id = kwargs.get('job_id', None)

        # for a json_ld citation from dataverse
        self.jsonld_citation = kwargs.get(KEY_JSONLD_CITATION, None)
        self.schema_info_dict = kwargs.get('SCHEMA_INFO_DICT', None)

        # for data source
        self.data_source_info = kwargs.get('data_source_info')

        self.celery_task = kwargs.get('celery_task')
        # to populate
        self.variable_info = {}  # { variable_name: ColumnInfo, ...}
        self.num_vars = None
        self.num_vars_complete = None

        # for error handling
        self.has_error = False
        self.error_message = None
        time_stamp = time.time()
        self.current_time = datetime.datetime.fromtimestamp(time_stamp)\
            .strftime('%Y-%m-%d %H:%M:%S')
        self.preprocess_id = None


        self.run_preprocess()

    def add_error_message(self, err_msg):
        """Add error message"""
        # print(err_msg)
        self.has_error = True
        self.error_message = err_msg

    def update_task_status(self):
        """Optional: update a celery task status"""
        if not self.celery_task:
            # no task available
            return

        if self.num_vars and self.num_vars > 0:
            self.celery_task.update_state(
                state='PROGRESS',
                meta={'current': self.num_vars_complete,
                      'total': self.num_vars})
            return

        self.add_error_message('Celery update failed')

    def run_preprocess(self):
        """Run preprocess"""
        if not isinstance(self.data_frame, pd.DataFrame):
            self.add_error_message('The dataframe is not valid')
            return False

        if not self.calculate_features():
            return False

        return True

    @staticmethod
    def load_from_file(input_file, **kwargs):
        """decide type/format, and name"""
        # Use new class to decide "csv file", "tab file"
        # etc
        file_format_util = FileFormatUtil(input_file, **kwargs)

        if file_format_util.has_error:
            return err_resp(file_format_util.error_message)
        else:
            if 'data_source_info' not in kwargs:
                kwargs['data_source_info'] = file_format_util.data_source_info

            runner = PreprocessRunner(\
                        file_format_util.dataframe,
                        **kwargs)
                        #job_id=job_id,
                        #data_source_info=file_format_util.data_source_info)
            if runner.has_error:
                return err_resp(runner.error_message)

            return ok_resp(runner)

    @staticmethod
    def load_update_file(preprocess_input, update_input):
        """ Returns the dict from the file input"""
        if not isfile(preprocess_input):
            return None, 'The file was not found: [%s]' % preprocess_input
        elif not isfile(update_input):
            return None, 'The file was not found: [%s]' % update_input

        filesize_preprocess = os.stat(preprocess_input).st_size
        filesize_update = os.stat(update_input).st_size
        if filesize_preprocess == 0:
            return None, 'The file size is zero: [%s]' % preprocess_input

        elif filesize_update == 0:
            return None, 'The file size is zero: [%s]' % update_input

        try:
            preprocess_input_dict = json.loads(preprocess_input, object_pairs_hook=OrderedDict)
        except TypeError as err_obj:
            err_msg = ('Input does not have Ordered dict convertable type'
                       '\n - File: %s\n - %s') % (preprocess_input, err_obj)
            return None, err_msg
        except Exception as err_obj:
            err_msg = ('Failed to convert into Orderd dict'
                       ' \n - File: %s\n - %s') % (preprocess_input, err_obj)
            return None, err_msg

        try:
            update_input_dict = json.loads(update_input,
                                           object_pairs_hook=OrderedDict,
                                           parse_float=decimal.Decimal)
        except TypeError as err_obj:
            err_msg = ('Input does not have Ordered dict convertable type'
                       '\n - File: %s\n - %s')% (update_input, err_obj)
            return None, err_msg
        except Exception as err_obj:
            err_msg = ('Failed to convert into Orderd dict'
                       ' \n - File: %s\n - %s') % (update_input, err_obj)
            return None, err_msg

        display_util = VariableDisplayUtil(preprocess_input_dict, update_input_dict)

        if display_util.has_error:
            return None, display_util.get_error_messages()

        return display_util, None

    def calculate_features(self):
        """For each variable, calculate summary stats"""
        if self.has_error:
            return False
        # Iterate through data frame and
        # run type guess, cal_stats, and plot_values on each ColumnInfo object
        #
        self.num_vars = len(self.data_frame.columns)
        self.num_vars_complete = 0

        for colnames in self.data_frame:
            # set stats for each column
            col_info = ColumnInfo(colnames)
            col_series = self.data_frame[colnames]

            type_guess_util = TypeGuessUtil(col_series, col_info)
            if type_guess_util.has_error():
                self.add_error_message(type_guess_util.get_error_message())
                return

            SummaryStatsUtil(col_series, col_info)
            PlotValuesUtil(col_series, col_info)
            # VariableDisplayUtil(col_series, col_info)
            # assign object info to the variable_info
            #
            self.num_vars_complete += 1
            self.variable_info[colnames] = col_info

        # print("completed column", self.num_vars_complete)
        # print(" Number of variable ", self.num_vars)
        return True

    '''
    def add_plot_info(self):
        """For each variable, add plot information as needed"""
        if self.has_error:
            return False

        if not self.variable_info:
            self.add_error_message('Error encountered. self.variable_info not available')
            return False

        for col_name, col_info in self.variable_info.items():
            # set stats for each column
            PlotValuesUtil(self.data_frame, col_info)

        return True
    '''

    def get_self_section(self):
        """
        {
          "self": {
            "description": "TwoRavens metadata generated by  https://github.com/TwoRavens/raven-metadata-service",
            "created": "current timestamp", # job.modified when done via the web
            "preprocess_id": 45  # null for now; job.id when done via the web
            "version": 1 # subsequent updates will increment this number
        },
        """
        self_section = OrderedDict()
        self_section['description'] = \
               ('TwoRavens metadata generated by'
                ' https://github.com/TwoRavens/raven-metadata-service')
        self_section['created'] = self.current_time
        self_section[col_const.PREPROCESS_ID] = self.job_id
        self_section['version'] = 1

        if self.schema_info_dict:
            self_section['schema'] = self.schema_info_dict

        return self_section


    def get_dataset_level_info(self):
        """
        "dataset": {
       "row_cnt": 1000,
       "variable_cnt": 25
                }
        """
        dataset_level_info = DatasetLevelInfo(self.data_frame)
        if dataset_level_info.has_error:
            info_dict = OrderedDict()
            info_dict["error"] = dataset_level_info.error_messages
            return info_dict

        info_dict = dataset_level_info.final_output
        if self.data_source_info:
            info_dict[col_const.DATA_SOURCE_INFO] = self.data_source_info.as_dict()

        if self.jsonld_citation:
            info_dict[col_const.DATASET_CITATION] = self.jsonld_citation
            #info_dict[col_const.DATA_SOURCE_INFO][col_const.DATASET_CITATION] = \
            #    self.jsonld_citation

        return info_dict

    def show_final_info(self):
        """Print the final info to the screen"""
        if self.has_error:
            err_msg = 'An error occurred earlier in the process:\n%s' % \
                      self.error_message
            print(err_msg)
            return

        info_string = self.get_final_dict(as_string=True)

        print(info_string)

    def get_final_json_indented(self, indent=4):
        """Return the final variable info as a JSON string"""
        if self.has_error:
            err_msg = 'An error occurred earlier in the process:\n%s' % \
                      self.error_message
            print(err_msg)
            return

        return self.get_final_dict(as_string=True,
                                   indent=indent)

    def get_final_json(self, indent=None):
        """Return the final variable info as a JSON string"""
        if self.has_error:
            err_msg = 'An error occurred earlier in the process:\n%s' % \
                      self.error_message
            print(err_msg)
            return

        return self.get_final_dict(as_string=True,
                                   indent=indent)

    def get_final_dict(self, as_string=False, **kwargs):
        """Return the preprocess data as an OrderedDict"""
        if self.has_error:
            err_msg = 'An error occurred earlier in the process:\n%s' % \
                      self.error_message
            print(err_msg)
            return

        fmt_variable_info = OrderedDict()   # capture the variables section
        fmt_display_variable_info = OrderedDict()   # capture the variable_display section
        fmt_display_variable_info['editable'] = ColumnInfo.get_editable_column_labels()
        # Iterate through each column and pull variable + variable_display info
        #
        for col_name, col_info in self.variable_info.items():
            # col_info.print_values()
            fmt_variable_info[col_name] = col_info.as_dict()
            fmt_display_variable_info[col_name] = VariableDisplayUtil.get_default_settings()

        # Format the entire document
        #
        overall_dict = OrderedDict()

        overall_dict[col_const.SELF_SECTION_KEY] = self.get_self_section()  # add the "self" section

        # add the 'dataset' section
        overall_dict[col_const.DATASET_LEVEL_KEY] = self.get_dataset_level_info()

        overall_dict[col_const.VARIABLES_SECTION_KEY] = fmt_variable_info    # add "variables"

        overall_dict[col_const.VARIABLE_DISPLAY_SECTION_KEY] = fmt_display_variable_info

        if as_string:
            # Convert the OrderedDict to a JSON string
            #
            indent_level = kwargs.get('indent', 4)
            if indent_level is None:
                pass
            elif (not isinstance(indent_level, int)) or indent_level > 50:
                indent_level = 4

            return json.dumps(overall_dict,
                              indent=indent_level,
                              cls=NumpyJSONEncoder)

        # w/o this step, a regular json.dumps() fails on the returned dict
        #
        jstring = json.dumps(overall_dict,
                             cls=NumpyJSONEncoder)
        return json.loads(jstring,
                          object_pairs_hook=OrderedDict,
                          parse_float=decimal.Decimal)



'''Contants used in the preprocess logic and metadata output'''
PREPROCESS_ID = 'preprocessId'


SELF_SECTION_KEY = 'self'
VERSION_KEY = 'version'
VARIABLES_SECTION_KEY = 'variables'
VARIABLE_DISPLAY_SECTION_KEY = 'variableDisplay'

# -----------------------------
# dataset
# -----------------------------
DATASET_LEVEL_KEY = 'dataset'
DATASET_DESCRIPTION = 'description'
DATASET_UNIT_OF_ANALYSIS = 'unitOfAnalysis'
DATASET_STRUCTURE = 'structure' # wide or long
STRUCTURE_WIDE = 'wide'
STRUCTURE_LONG = 'long'

DATASET_ROW_CNT = 'rowCount'
DATASET_VARIABLE_CNT = 'variableCount'
DATASET_CITATION = 'citation'

DATA_SOURCE_INFO = 'dataSource'
DATA_SOURCE_NAME = 'name'
DATA_SOURCE_TYPE = 'type'
DATA_SOURCE_FORMAT = 'format'
DATA_SOURCE_FILESIZE = 'filesize'


UNKNOWN = 'unknown'
NOT_IMPLEMENTED = 'NOT IMPLEMENTED'
NOT_APPLICABLE = 'NA'

# --------------------------------------
# variable labels for the metadata file
# --------------------------------------
DESCRIPTION_LABEL = 'description'

# --------------------------------------
# numchar constants for metadata file
# --------------------------------------
NUMCHAR_LABEL = 'numchar'

NUMCHAR_NUMERIC = 'numeric'
NUMCHAR_CHARACTER = 'character'

NUMCHAR_VALUES = (NUMCHAR_NUMERIC,
                  NUMCHAR_CHARACTER)

# --------------------------------------
# interval constants for metadata file
# --------------------------------------
INTERVAL_CONTINUOUS = 'continuous'
INTERVAL_DISCRETE = 'discrete'
INTERVAL_VALUES = (INTERVAL_CONTINUOUS,
                   INTERVAL_DISCRETE)

# --------------------------------------
# nature constants for metadata file
# --------------------------------------
NATURE_LABEL = 'nature'

NATURE_NOMINAL = 'nominal'
NATURE_ORDINAL = 'ordinal'
NATURE_INTERVAL = 'interval'
NATURE_RATIO = 'ratio'
NATURE_PERCENT = 'percent'
NATURE_OTHER = 'other'
NATURE_VALUES = (NATURE_NOMINAL,
                 NATURE_ORDINAL,
                 NATURE_INTERVAL,
                 NATURE_RATIO,
                 NATURE_PERCENT,
                 NATURE_OTHER)

# --------------------------------------
# binary constants for metadata file
# --------------------------------------
BINARY_YES = True
BINARY_NO = False
BINARY_VALUES = (BINARY_YES,
                 BINARY_NO)

# --------------------------------------
# time constants for metadata file
# --------------------------------------
TIME_UNKNOWN = 'unknown'
TIME_YES = 'yes'
TIME_NO = 'no'
TIME_VALUES = (TIME_YES,
               TIME_NO)

# --------------------------------------
# plot types for metadata file
# --------------------------------------
PLOT_BAR = "bar"
PLOT_CONTINUOUS = "continuous"


CUSTOM_KEY = 'customStatistics'
UPDATE_VARIABLE_DISPLAY = 'UPDATE_VARIABLE_DISPLAY'
UPDATE_CUSTOM_STATISTICS = 'UPDATE_CUSTOM_STATISTICS'
UPDATE_TO_CUSTOM_STATISTICS = 'UPDATE_TO_CUSTOM_STATISTICS'
DELETE_CUSTOM_STATISTICS = 'DELETE_CUSTOM_STATISTICS'



""" Module for type guessing """
import pandas as pd
from pandas.api.types import is_float_dtype, is_numeric_dtype

import raven_preprocess.col_info_constants as col_const
from raven_preprocess.basic_utils.basic_err_check import BasicErrCheck


class TypeGuessUtil(BasicErrCheck):
    """Check variable types of a dataframe"""
    def __init__(self, col_series, col_info):
        """Init with a pandas dataframe"""
        assert col_series is not None, "dataframe can't be None"

        self.col_series = col_series
        self.col_info = col_info
        self.binary = False
        # # final outout returned
        self.check_types()

    def check_types(self):
        """check the types of the dataframe"""
        # assert self.colnames, 'self.colnames must have values'
        # number of missing entries
        #
        self.col_info.invalid = int(self.col_series.isnull().sum())

        # number of valid entries
        #
        self.col_info.valid = int(self.col_series.count())

        # set time , what exactly we want to do with this
        #
        self.col_info.time_val = self.check_time(self.col_series)

        # Drop nulls...
        self.col_series.dropna(inplace=True)

        uniques = self.col_series.unique()

        binary = self.check_binary(len(uniques))

        # set up binary values..
        if binary:
            self.col_info.binary = col_const.BINARY_YES
        else:
            self.col_info.binary = col_const.BINARY_NO
        if self.is_not_numeric(self.col_series) or self.is_logical(self.col_series):

            self.col_info.numchar_val = col_const.NUMCHAR_CHARACTER
            self.col_info.default_interval = col_const.INTERVAL_DISCRETE
            self.col_info.nature = col_const.NATURE_NOMINAL

        else:

            try:
                series_info = self.col_series.astype('int')
            except ValueError as err_obj:
                user_msg = ('Type guess error when converting'
                            ' to int: %s') % err_obj
                self.add_err_msg(user_msg)
                return

            if any(series_info.isnull()):
                # CANNOT REACH HERE B/C NULLS ARE DROPPED!
                #
                self.col_info.numchar_val = col_const.NUMCHAR_CHARACTER
                self.col_info.nature = col_const.NATURE_NOMINAL
                self.col_info.default_interval = col_const.INTERVAL_DISCRETE
            else:
                self.col_info.numchar_val = col_const.NUMCHAR_NUMERIC

                if is_float_dtype(series_info):
                    self.col_info.default_interval = col_const.INTERVAL_CONTINUOUS
                    self.col_info.nature = self.check_nature(series_info, True)

                else:
                    self.col_info.default_interval = col_const.INTERVAL_DISCRETE
                    self.col_info.nature = self.check_nature(series_info, False)




    @staticmethod
    def is_not_numeric(var_series):
        """Check if pandas Series is a numeric"""
        assert isinstance(var_series, pd.Series), \
            "var_series must be a pandas.Series. Found type: (%s)" % type(var_series)

        var_series.dropna(inplace=True)
        if var_series.size == 0:
            # print("character")
            return True
        elif var_series.dtype == 'bool':
            return True

        if is_numeric_dtype(var_series):
            return False
        else:
            return True

    @staticmethod
    def is_logical(var_series):
        """Check if pandas Series contains boolean values"""
        assert isinstance(var_series, pd.Series), \
            "var_series must be a pandas.Series. Found type: (%s)" % type(var_series)

        var_series.dropna(inplace=True)

        # Check the dtype
        #    "bool" - True, clearly logical
        #    "object" - possibly logical that had contained np.Nan
        #    ~anything else~ - False
        #
        if var_series.dtype == 'bool':
            return True
        elif var_series.dtype != 'object':
            return False

        # It's an object.  Check if all the values
        #   either True or False
        #
        total = var_series.size
        total_cnt = 0
        for val, cnt in var_series.value_counts().iteritems():
            if val is True or val is False:
                total_cnt = total_cnt + cnt

        if total_cnt == total:
            # This is a boolean -- everything was either True or False
            #
            return True

        return False

    @staticmethod
    def check_nature(data_series, continuous_check):
        """Check the nature of the Series"""
        if continuous_check:
            if data_series.between(0, 1).all():
                return col_const.NATURE_PERCENT
            elif data_series.between(0, 100).all() and min(data_series) < 15 and max(data_series) > 85:
                return col_const.NATURE_PERCENT
            else:
                return col_const.NATURE_RATIO

        else:
            return col_const.NATURE_ORDINAL

    @staticmethod
    def check_time(var_series):
        """Unimplemented"""
        assert isinstance(var_series, pd.Series), \
            "var_series must be a pandas.Series. Found type: (%s)" % type(var_series)

        return col_const.UNKNOWN

    @staticmethod
    def check_binary(unique_size):
        """ check if the series is binary or not """
        if unique_size is 2:
            return True
        else:
            return False



"""Convenience methods for printing to screen"""
import sys


def msg(user_msg):
    """Print"""
    print(user_msg)


def dashes(char='-'):
    """Dashed line"""
    msg(40*char)


def msgt(user_msg):
    """print message with dashed line before/after"""
    dashes()
    msg(user_msg)
    dashes()


def msgn(user_msg):
    """print message with dashed line before"""
    dashes()
    msg(user_msg)


def msgx(user_msg):
    """Print message and exit program--hard exit"""
    dashes('=')
    msg(user_msg)
    dashes('=')
    sys.exit(0)



"""Constants related to update metadata requests
Will be switched to a JSON schema check

Example update request:
 {
    "preprocessId": 5,
    "variableUpdates": {
       "cylinders" : {
         "viewable": true,
         "omit": ["mean", "median"],
         "valueUpdates": {
             "numchar":"character",
             "nature": "ordinal"
         }
       },
       "mpg": {
         "viewable": false,
         "omit": [],
         "valueUpdates": {

         }
       }
    }
}
"""

VARIABLE_UPDATES = 'variableUpdates'

VIEWABLE_KEY = 'viewable'
OMIT_KEY = 'omit'
VALUE_UPDATES_KEY = 'valueUpdates'

# ----------------
START_ROW = 'startRow'
NUM_ROWS = 'numberRows'



from datetime import date, datetime
import simplejson as json
import numpy as np


class NumpyJSONEncoder(json.JSONEncoder):
    """class to encode the data"""
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        else:
            return super(NumpyJSONEncoder, self).default(obj)


# Encoder function
def np_dumps(obj):
    return json.dumps(obj, cls=NumpyJSONEncoder)



from __future__ import print_function
from collections import OrderedDict

import raven_preprocess.col_info_constants as col_const

SOURCE_TYPE_FILE = 'File'

class DataSourceInfo(object):

    def __init__(self, name, source_type, **kwargs):
        """Set data source"""
        self.name = name
        self.source_type = source_type

        self.source_format = kwargs.get('source_format')
        self.filesize = kwargs.get('filesize')

    def as_dict(self):
        """Return DataSourceInfo as an OrderedDict()"""
        data = OrderedDict()
        data[col_const.DATA_SOURCE_NAME] = self.name
        data[col_const.DATA_SOURCE_TYPE] = self.source_type

        # may be None in final output
        data[col_const.DATA_SOURCE_FORMAT] = self.source_format

        # optional
        if self.filesize:
            data[col_const.DATA_SOURCE_FILESIZE] = self.filesize

        return data



""" Module for plot values eg: type, cdf, labl etc"""
from __future__ import print_function
import logging
import numpy as np
import pandas as pd
from scipy import stats

from raven_preprocess.column_info import ColumnInfo
import raven_preprocess.col_info_constants as col_const
logger = logging.getLogger(__name__)


class PlotValuesUtil(object):
    """ Class to set up for plot values eg: type, cdf, labl etc"""
    def __init__(self, col_series, col_info):
        assert isinstance(col_series, pd.Series), "col_series must be a pandas.Series object"
        assert isinstance(col_info, ColumnInfo), "col_info must be a ColumnInfo object"

        self.plot_values = list()
        logger.debug("plot values time")
        #self.dataframe = dataframe
        self.col_info = col_info
        self.colname = self.col_info.colname
        self.col_series = col_series
        self.histlimit = 13
        self.output = {}
        self.cdfx = None
        self.cdfy = None

        self.cal_plot_values()
        self.col_info.tidy_plot_values()



    def ecdf(self, data):
        """Compute ECDF for a one-dimensional array of measurements."""
        # Number of data points: size
        raw_data = np.array(self.col_series)

        # x-data for the ECDF: x_
        x_value = np.sort(data)
        size_data = raw_data.size
        # y-data for the ECDF: y
        y_value = []

        for i in x_value:
            temp = raw_data[raw_data <= i]
            val = temp.size / size_data
            y_value.append(val)

        return y_value

    def cal_plot_values(self):
        """Compute all plot values."""

        nat = self.col_info.nature
        my_interval = self.col_info.interval
        self.col_series.dropna(inplace=True)
        if col_const.NATURE_NOMINAL != nat:
            logger.debug("into not nominal %s", self.colname)

            uniques = np.sort(self.col_series.unique())
            lu = len(uniques)

            if lu < self.histlimit:
                logger.debug("into it %s", self.colname)
                # code for plot values
                self.col_info.plot_type = col_const.PLOT_BAR

                for val, cnt in self.col_series.sort_values().value_counts().items():
                    if type(val) is not str:
                        try:
                            self.output[str(val)] = cnt
                        except TypeError:
                            try:
                                self.output[repr(val)] = cnt
                            except TypeError:
                                pass
                self.col_info.plot_values = self.output

                # code for cdf values
                self.cdfx = np.sort(uniques)
                self.col_info.cdf_plot_type = col_const.PLOT_BAR
                self.col_info.cdf_plotx = np.linspace(start=min(self.cdfx),
                                                      stop=max(self.cdfx), num=len(self.cdfx))
                self.col_info.cdf_ploty = self.ecdf(self.col_info.cdf_plotx)

            else:
                # code for plot values
                self.col_info.plot_type = col_const.PLOT_CONTINUOUS
                # here the code for plotx and ploty comes using r density function
                if self.col_series is not None:
                    kernel = stats.gaussian_kde(self.col_series)
                    x = np.linspace(start=min(self.col_series) - kernel.factor,
                                    stop=max(self.col_series) + kernel.factor, num=50)
                    self.col_info.plotx = x
                    self.col_info.ploty = kernel(x)

                # code for cdf values
                self.col_info.cdf_plot_type = col_const.PLOT_CONTINUOUS
                if lu >= 50 or (lu < 50 and my_interval != col_const.INTERVAL_DISCRETE):
                    self.col_info.cdf_plotx = np.linspace(start=min(self.col_series),
                                                          stop=max(self.col_series), num=50)

                else:
                    self.col_info.cdf_plotx = np.linspace(start=min(self.col_series),
                                                          stop=max(self.col_series), num=lu)
                self.col_info.cdf_ploty = self.ecdf(self.col_info.cdf_plotx)

        else:
            """Here data nature is not nominal"""
            logger.debug("into it *******%s", self.colname)
            # code for plot values
            self.col_info.plot_type = col_const.PLOT_BAR

            for val, cnt in self.col_series.sort_values().value_counts().items():
                if not isinstance(val, str):
                    try:
                        self.output[str(val)] = cnt
                    except TypeError:
                        try:
                            self.output[repr(val)] = cnt
                        except TypeError:
                            pass
                    del cnt
                else:
                    self.output[str(val)] = cnt

            self.col_info.plot_values = self.output

            # code for cdf values
            self.col_info.cdf_plot_type = None
            self.col_info.cdf_plotx = None
            self.col_info.cdf_ploty = None

        # if metadataflag  !=1
        self.col_info.labl = ""



from __future__ import print_function
""" this is the information of dataset level info"""
from collections import OrderedDict

import raven_preprocess.col_info_constants as col_const


class DatasetLevelInfo(object):

    def __init__(self, df):
        """This class sets the dataset level info of the preprocess file. """

        self.dataframe = df
        self.rows_count = None
        self.variables_count = None
        self.has_error = False
        self.error_messages = []
        self.final_output = OrderedDict()

        self.set_values()

    def set_values(self):
        """
        "dataset": {
       "row_cnt": 1000,
       "variable_cnt": 25
            }
        """
        if self.dataframe is not None:
            self.rows_count = self.dataframe.shape[0]  \
                # shape[0] gives the number of records/rows and is faster then count
            self.variables_count = len(self.dataframe.columns)
        else:
            self.has_error = True
            self.error_messages.append(" There is no data available to get dataset level info")
            return
        if self.rows_count < 1:
            self.has_error = True
            self.error_messages.append(" This is an empty dataframe with no record")
            return
        if self.variables_count < 1:
            self.has_error = True
            self.error_messages.append(" This is an empty dataframe with no variables")
            return

        self.final_output[col_const.DATASET_DESCRIPTION] = ''
        self.final_output[col_const.DATASET_UNIT_OF_ANALYSIS] = ''
        self.final_output[col_const.DATASET_STRUCTURE] = col_const.STRUCTURE_LONG
        self.final_output[col_const.DATASET_ROW_CNT] = self.rows_count
        self.final_output[col_const.DATASET_VARIABLE_CNT] = self.variables_count



""" File to get the updated metadata file with custom statistics"""
from collections import OrderedDict
import re
import json
import raven_preprocess.col_info_constants as col_const
from raven_preprocess.column_info import ColumnInfo
from raven_preprocess.np_json_encoder import NumpyJSONEncoder
from raven_preprocess.version_number_util import VersionNumberUtil

ALL_VARIABLE_ATTRIBUTES = [x[0] for x in ColumnInfo.get_variable_labels()]
VIEWABLE_VALUES = [True, False]
CUSTOM_STATISTICS_NAME = 'name'
CUSTOM_STATISTICS_VARIABLES = 'variables'
CUSTOM_STATISTICS_IMAGE = 'image'
CUSTOM_STATISTICS_DESCRIPTION = 'description'
CUSTOM_STATISTICS_REPLICATION = 'replication'
CUSTOM_STATISTICS_VIEWABLE = 'viewable'
CUSTOM_STATISTICS_VALUE = 'value'
CUSTOM_STATISTICS_DELETE = 'delete'
CUSTOM_STATISTICS_UPDATE = 'updates'
CUSTOM_STATISTICS_ID = 'id'


class CustomStatisticsUtil(object):
    def __init__(self, preprocess_json, custom_statistics_json):
        """class for the custom statistics process"""

        assert isinstance(preprocess_json, dict),\
            "preprocess_json must be a dict/OrderedDict"

        self.preprocess_json = preprocess_json
        self.custom_statistics_json = custom_statistics_json

        # for error handling
        self.has_error = False
        self.error_messages = []
        self.original_json = None

        # call the statistic function

    def add_error_message(self, err_msg):
        """Add error message"""
        # print(err_msg)
        self.has_error = True
        self.error_messages.append(err_msg)

    def get_updated_metadata(self, as_string=False):
        """Return the modified metadata--which is in the 'original_json' """
        assert self.has_error is False, \
            "Make sure that 'has_error' is False before using this method"

        if as_string:
            return json.dumps(self.original_json,
                              indent=4,
                              cls=NumpyJSONEncoder)

        return self.original_json

    def get_error_messages(self):
        """Return the list of error messages"""
        print("Error messages ", self.error_messages)
        return self.error_messages

    def custom_statistics_create_id(self):   # need to think of generating id
        if col_const.CUSTOM_KEY not in self.preprocess_json:
            var_id = 'id_000001'
        else:
            ids = list(self.preprocess_json[col_const.CUSTOM_KEY])
            ids.sort()
            latest_id = ids[-1]
            _unused, idnum = latest_id.split('_')
            next_num = str(int(idnum) + 1)
            var_id = 'id_%s' % next_num.zfill(6)
        return var_id

    def custom_statistics_check_name(self, name):
        statistics_name = name
        if not re.match(r'^[A-Za-z0-9_]+$', statistics_name):   # check if the name is alpha numerics , \
                                            #  i.e does not contain special characters or empty spaces
            self.add_error_message('The name is not valid (Should be alpha-numeric)')
        # if col_const.CUSTOM_KEY in self.preprocess_json:
            # for val in self.preprocess_json[col_const.CUSTOM_KEY]:
            #     print("Value coming ",self.preprocess_json[col_const.CUSTOM_KEY][val]['name'])
            #     if name == self.preprocess_json[col_const.CUSTOM_KEY][val]['name']:
            #         print("******** error **********")
            #         self.add_error_message('The variable name %s already exist in the metadata file' % name)
        return statistics_name

    def custom_statistics_check_variables(self, var_list, variables):
        # print("all var ", var_list)
        for var in variables:
            if var not in var_list:
                self.add_error_message('The variable %s does not exist in the metadata file' % var)

        return variables

    def custom_statistics_check_image(self, image_url):
        if image_url is None:
            self.add_error_message('image url not found!')

        return image_url

    @staticmethod
    def is_major_update():
        """add or change a custom stat"""
        return False

    @staticmethod
    def is_minor_update():
        """add or change a custom stat"""
        return True

    def custom_statistics_check_value(self, value):
        if value is None:
            self.add_error_message(' value not found ')

        return value

    def custom_statistics_check_description(self, desc):
        if desc is None:
            self.add_error_message(' description not found!')
            return

        return desc

    def custom_statistics_check_replication(self, rep):
        if rep is None:
            self.add_error_message(' replication not found! ')
            return

        return rep

    def custom_statistics_check_viewable(self, viewable):
        if viewable not in VIEWABLE_VALUES:
            self.add_error_message('Viewable should be either True or False')
            return
        return viewable

    def custom_statistics_update(self):
        """Main function for appending the data"""
        """
               sample Update json coming :
               {
          "preprocess_id":1677,
          "custom_statistics":[
             {
                "name":"Third order statistic",
                "variables":"lpop,bebop",
                "image":"http://www.google.com",
                "value":23.45,
                "description":"Third smallest value",
                "replication":"sorted(X)[2]",
                "omit":false
             },
             {
                "name":"Fourth order statistic",
                "variables":"pop,bebop",
                "image":"http://www.youtube.com",
                "value":29.45,
                "description":"Fourth smallest value",
                "replication":"sorted(X)[3]",
                "omit":false
             }
          ]
       }
               """
        # preprocess_id = self.custom_statistics_json['preprocess_id']
        var_list = list(self.preprocess_json[CUSTOM_STATISTICS_VARIABLES])

        for dat in self.custom_statistics_json:
            # print("dat ", dat)
            if CUSTOM_STATISTICS_NAME in dat:
                name = self.custom_statistics_check_name(dat[CUSTOM_STATISTICS_NAME])   # required = True
            else:
                name = ''
                self.add_error_message(' name is required')
            # id = self.custom_statistics_create_id(name)
            if CUSTOM_STATISTICS_VARIABLES in dat:

                variables = self.custom_statistics_check_variables(var_list,
                                                                   dat[CUSTOM_STATISTICS_VARIABLES])   # required = True
            else:
                variables = ''
                self.add_error_message(' at least one variable is required')

            if CUSTOM_STATISTICS_IMAGE in dat:
                image = self.custom_statistics_check_image(dat[CUSTOM_STATISTICS_IMAGE])
            else:
                image = []

            if CUSTOM_STATISTICS_VALUE:
                value = self.custom_statistics_check_value(dat[CUSTOM_STATISTICS_VALUE])    # required = True
            else:
                value = ''
                self.add_error_message(' value is required')

            if CUSTOM_STATISTICS_DESCRIPTION in dat:
                description = self.custom_statistics_check_description(dat[CUSTOM_STATISTICS_DESCRIPTION])
            else:
                description = None

            if CUSTOM_STATISTICS_REPLICATION in dat:
                replication = self.custom_statistics_check_replication(dat[CUSTOM_STATISTICS_REPLICATION])
            else:
                replication = None

            if CUSTOM_STATISTICS_VIEWABLE in dat:
                viewable = self.custom_statistics_check_viewable(dat[CUSTOM_STATISTICS_VIEWABLE])
            else:
                viewable = True  # default
            if self.has_error:
                return False
            else:
                data = {
                        CUSTOM_STATISTICS_NAME: name,
                        CUSTOM_STATISTICS_VARIABLES: variables,
                        CUSTOM_STATISTICS_IMAGE: image,
                        CUSTOM_STATISTICS_VALUE: value,
                        CUSTOM_STATISTICS_DESCRIPTION: description,
                        CUSTOM_STATISTICS_REPLICATION: replication,
                        "display": {
                            CUSTOM_STATISTICS_VIEWABLE: viewable
                        }

                    }

                self.add_to_original(data)

        self.original_json = OrderedDict(self.preprocess_json)

        success, updated_or_err = VersionNumberUtil.update_version_number(
                                        self.original_json,
                                        self.is_major_update())


        if not success:
            self.add_error_message(updated_or_err)
            return False

    def add_to_original(self, data):
        id_name = self.custom_statistics_create_id()
        # self.original_json= self.preprocess_json
        if col_const.CUSTOM_KEY not in self.preprocess_json:
            self.preprocess_json[col_const.CUSTOM_KEY] = {id_name: data}
        else:
            self.preprocess_json[col_const.CUSTOM_KEY][id_name] = data

        # print(self.original_json)

    # update functions for custom_stats_update

    def check_ids(self, id_name):
        """check for id"""
        id_list = list(self.preprocess_json[col_const.CUSTOM_KEY])    # return list of ids
        # print("id name and id list ", id_name, id_list)
        if id_name in id_list:
            # print(' id %s is there ' % id_name)
            return True, None
        else:
            return False, ' id %s not found in requested update' % id_name

    def make_update(self, id_name, update_json):
        """ make changes to preprocess json"""
        for val in update_json:
            # print(" this is val ", val)
            # print("this is the updateing object ", self.preprocess_json['custom_statistics'][id])
            try:
                self.preprocess_json[col_const.CUSTOM_KEY][id_name][val] = update_json[val]
            except KeyError:
                self.add_error_message('%s not present in the preprocess file ' % val)
                return False

    def basic_update_structure_check(self, update_json, type_input):
        """ check if all updates has update and ids"""
        search = None
        if type_input == CUSTOM_STATISTICS_DELETE:
            search = CUSTOM_STATISTICS_DELETE
        elif type_input == CUSTOM_STATISTICS_UPDATE:
            search = CUSTOM_STATISTICS_UPDATE
        if CUSTOM_STATISTICS_ID not in update_json:
            # self.add_error_message('id section is not present in request')
            return False, 'id section is not present in request'
        if search not in update_json:
            # self.add_error_message('%s section not present in request' % search)
            return False, '%s section not present in request' % search

        return True, None

    def update_custom_stats(self):
        """ The update is done here"""
        """ Sample update_json
                 [
            {
              "id": "id_1",
              "updates": {
                "name": "Fourth order statistic",
                "value": 40
              }
            },
            {
              "id": "id_2"
              "updates": {
                "name": "This will be a new statistic",
                "value": 40
              }
            }
          ]
        """
        # going through each update
        for update in self.custom_statistics_json:
            check, msg = self.basic_update_structure_check(update, CUSTOM_STATISTICS_UPDATE)
            if check is False:
                self.add_error_message(msg)

            id_check, msg = self.check_ids(update[CUSTOM_STATISTICS_ID])
            if id_check is False:
                self.add_error_message(msg)

            if not self.has_error:
                self.make_update(update[CUSTOM_STATISTICS_ID], update[CUSTOM_STATISTICS_UPDATE])

        self.original_json = OrderedDict(self.preprocess_json)
        success, updated_or_err = VersionNumberUtil.update_version_number(self.original_json, self.is_major_update())
        if not success:
            self.add_error_message(updated_or_err)
            return False
        # print("updates to custom_statistics : ", self.original_json)

    def make_deletion(self, id_num, update):
        """ here we delete"""
        if CUSTOM_STATISTICS_ID in update:
            del self.preprocess_json[col_const.CUSTOM_KEY][id_num]
        else:
            for val in update:
                try:
                    del self.preprocess_json[col_const.CUSTOM_KEY][id_num][val]
                except KeyError:
                    self.add_error_message('%s not present in file' % val)
                    return False

    def delete_custom_stat(self):
        """ delete fields/custom stats"""
        # print("custom json ", self.custom_statistics_json)
        for delete_obj in self.custom_statistics_json:
            # print("this is delete obj ", delete_obj)
            check, msg = self.basic_update_structure_check(delete_obj, CUSTOM_STATISTICS_DELETE)
            if not check:
                self.add_error_message(msg)

            id_check, msg = self.check_ids(delete_obj[CUSTOM_STATISTICS_ID])
            if id_check is False:
                self.add_error_message(msg)

            if not self.has_error:
                self.make_deletion(delete_obj[CUSTOM_STATISTICS_ID], delete_obj[CUSTOM_STATISTICS_DELETE])

        self.original_json = OrderedDict(self.preprocess_json)

        # print("After deletion ", self.original_json)

        success, updated_or_err = VersionNumberUtil.update_version_number(self.original_json,
                                                                          self.is_major_update())
        if not success:
            self.add_error_message(updated_or_err)
            return False



"""
An object with some basic methods for capturing errors
"""

class BasicErrCheck(object):

    error_found = False
    error_message = None

    def has_error(self):
        """Did an error occur?"""
        return self.error_found

    def get_error_message(self):
        """Return the error message if 'has_error' is True"""
        assert self.has_error(),\
            "Please check that '.has_error()' is True before using this method"

        return self.error_message

    def add_err_msg(self, err_msg):
        """Add an error message"""
        self.error_found = True
        self.error_message = err_msg

def try_it():
    """quick check"""
    b = BasicErrCheck()
    print('error_found', b.error_found)
    print('error_message', b.error_message)
    print(b.has_error())
    b.add_err_msg('uh oh')
    print(b.has_error())
    print(b.get_error_message())

if __name__ == '__main__':
    try_it()



"""Main module to call all sub modules"""
from os.path import abspath, dirname, join
import logging
import sys

CURRENT_DIR = dirname(abspath(__file__))
sys.path.append(dirname(CURRENT_DIR))

from raven_preprocess.preprocess_runner import PreprocessRunner
from raven_preprocess.basic_utils.basic_response import (ok_resp, err_resp)
from raven_preprocess.msg_util import msgt

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)

INPUT_DIR = join(dirname(CURRENT_DIR), 'input')
OUTPUT_DIR = join(dirname(CURRENT_DIR), 'output')

def run_preprocess(input_file, output_filepath=None):
    """Main test run class for this module"""
    run_info = PreprocessRunner.load_from_file(input_file)

    if not run_info.success:
        msgt(run_info.err_msg)
        #print(err_resp(err_msg))
        return

    runner = run_info.result_obj

    runner.show_final_info()
    #return ok_resp(runner.get_final_json(indent=4))

    jstring = runner.get_final_json(indent=4)

    if output_filepath:
        try:
            open(output_filepath, 'w').write(jstring)
            msgt('file written: %s' % output_filepath)
        except OSError as os_err:
            msgt('Failed to write file: %s' % os_err)


def show_instructions():
    """show command line instructions"""
    info = """
--------------------------
preprocess a single file
--------------------------

# write output to screen
> python preprocess.py [input csv file]

# write output to screen and file
> python preprocess.py [input csv file] [output file name]

OR

# test input and output files
> python preprocess.py test

"""
    print(info)

if __name__ == '__main__':

    if len(sys.argv) == 2:
        if sys.argv[1] == 'test':
            input_csv = join(INPUT_DIR, 'test_file_01.csv')
            output_file = join(OUTPUT_DIR, 'test_file_01_preprocess.json')
            run_preprocess(input_csv, output_file)
        else:
            run_preprocess(sys.argv[1])

    elif len(sys.argv) == 3:
        run_preprocess(sys.argv[1], sys.argv[2])
    else:
        show_instructions()



"""Module for preprocess structure"""

from collections import OrderedDict
import json
from raven_preprocess.np_json_encoder import NumpyJSONEncoder
import raven_preprocess.col_info_constants as col_const


class ColumnInfo(object):
    """Sets up the expected structure of the whole preprocess output file"""

    # names of ColumnInfo *attributes*, not labels, that are editable
    EDITABLE_COLUMNS = ['labl', 'numchar_val', 'nature', 'time_val']

    def __init__(self, colname):
        """Init with column name"""
        # -------------
        # self data
        # -------------

        self.schema = None
        self.description = None
        self.created = None
        self.preprocess_id = None
        self.data_url = None
        self.format = None
        self.preprocess_version = None
        self.schema_version = None

        # -------------
        # more general
        # -------------
        self.colname = colname
        self.labl = None

        self.valid = None
        self.invalid = None

        # ----------------------
        # Type Guess Info
        # ----------------------
        self.numchar_val = None
        self.default_interval = None
        self.nature = None
        self.time_val = None
        self.binary = None

        # set at type util
        self.varnames_sum_stat = None

        # Stats Info
        self.mode = []
        self.freqmode = None
        self.mid = None
        self.fewest = []
        self.freqmid = None
        self.freqfewest = None
        self.median = None
        self.mean = None
        self.max = None
        self.min = None
        self.std_dev = None
        self.herfindahl = None

        self.uniques = None

        self.numchar = None
        self.nature = None
        self.binary = None
        self.interval = None
        self.time = None


        #plot vlaues
        self.plot_values = {}
        self.plot_type = None
        self.plotx = None
        self.ploty = None
        self.cdf_plot_type = None
        self.cdf_plotx = None
        self.cdf_ploty = None

    def is_numeric(self):
        # is this NUMCHAR_NUMERIC?
        return self.numchar_val == col_const.NUMCHAR_NUMERIC

    def is_character(self):
        # is this NUMCHAR_CHARACTER?
        return self.numchar_val == col_const.NUMCHAR_CHARACTER

    @staticmethod
    def is_valid_nature(val):
        """Check if the nature is valid"""
        return val in col_const.NATURE_VALUES

    @staticmethod
    def is_valid_numchar(val):
        """Check if the nature is valid"""
        return val in col_const.NUMCHAR_VALUES


    def get_numeric_attribute_names(self):
        """These attributes, when set, are always numeric.  Ex/ mean, median, etc."""

        return  ('invalid', 'valid', 'uniques',
                 'median', 'mean', 'max', 'min',
                 'freqmode', 'freqfewest', 'freqmid',
                 'std_dev', 'herfindahl',
                 'plot_values', 'plotx', 'ploty',
                 'cdf_plotx', 'cdf_ploty')

    def is_numeric_attribute(self, ye_attr_name):
        """Test if the attribute is numeric (or null if not set)"""
        if ye_attr_name in self.get_numeric_attribute_names():
            return True
        return False


    def set_fewest(self, fewest_list):
        """Take up to 5 values that occur the fewest number of times"""
        if fewest_list is None:
            self.fewest = []

        self.fewest = fewest_list[:5]

    def set_mode(self, mode_list):
        """Take up to 5 values that occur the greatest number of times"""
        if mode_list is None:
            self.mode = []

        self.mode = mode_list[:5]

    @staticmethod
    def get_editable_column_labels():
        """Return the labels associated with the editable columns"""
        lookup = ColumnInfo.get_variable_label_lookup()

        editable_list = [lookup.get(varname) for varname in ColumnInfo.EDITABLE_COLUMNS]

        assert not None in editable_list,\
            ('SERIOUS ERROR!! In ColumnInfo, EDITABLE_COLUMNS'
             ' has an attribute that either does not exist or'
             ' doesn\'t have a label')

        return editable_list


    @staticmethod
    def get_variable_label_lookup():
        """Return a lookup consisting of {"variable name": "label name"}
        Exclude variables starting with "default" as in "defaultTime" """

        return {info[1]: info[0]
                for info in ColumnInfo.get_variable_labels()
                if not info[0].startswith('default')}


    @staticmethod
    def get_variable_labels():
        """Set labels for variable output.  List of:

         (label, variable name)

        Example of iterating through to show labels and values:
            ```
            for label, varname in self.get_variable_labels():
                variable_val = self.__dict__.get(varname)
                print('%s: %s' % (label, variable_val))
            ```
        """
        label_list = (
            ('variableName', 'colname'),
            (col_const.DESCRIPTION_LABEL, 'labl'),

            (col_const.NUMCHAR_LABEL, 'numchar_val'),
            (col_const.NATURE_LABEL, 'nature'),

            ('binary', 'binary'),
            ('interval', 'default_interval'),
            ('time', 'time_val'),

            ('invalidCount', 'invalid'),
            ('validCount', 'valid'),
            ('uniqueCount', 'uniques'),

            ('median', 'median'),
            ('mean', 'mean'),
            ('max', 'max'),
            ('min', 'min'),

            ('mode', 'mode'),
            ('modeFreq', 'freqmode'),
            ('fewestValues', 'fewest'),
            ('fewestFreq', 'freqfewest'),
            ('midpoint', 'mid'),
            ('midpointFreq', 'freqmid'),

            ('stdDev', 'std_dev'),
            ('herfindahlIndex', 'herfindahl'),

            ('plotValues', 'plot_values'),
            ('pdfPlotType', 'plot_type'),
            ('pdfPlotX', 'plotx'),
            ('pdfPlotY', 'ploty'),
            ('cdfPlotType', 'cdf_plot_type'),
            ('cdfPlotX', 'cdf_plotx'),
            ('cdfPlotY', 'cdf_ploty'),

            )
        # print("-"*20)
        # print(label_list)
        return label_list


    def tidy_plot_values(self):
        """for plot values that are None, set them to None or "NA" """
        attr_names = ('plot_type', 'plotx', 'ploty',
                      'cdf_plot_type', 'cdf_plotx', 'cdf_ploty')

        for pname in attr_names:
            val = self.__dict__.get(pname)
            if val in ('', []):
                self.__dict__[pname] = None #col_const.NOT_APPLICABLE

        # Next 2 lines need to be fixed in the plot
        # value calculation section

        if self.plotx is None and self.ploty is None:
            self.plot_type = None

        if self.cdf_plotx is None and self.cdf_ploty is None:
            self.cdf_plot_type = None

        # Until mid is updated
        self.mid = None
        self.freqmid = None

    def print_values(self):
        """print to screen"""
        print('---- %s ----' % self.colname)
        for label, varname in self.get_variable_labels():
            print('%s: %s' % (label, self.__dict__.get(varname)))

    def as_dict(self, as_string=False):
        """For final output"""
        ordered_dict = OrderedDict()

        for label, varname in self.get_variable_labels():
            ordered_dict[label] = self.__dict__.get(varname)

        if as_string:
            return json.dumps(ordered_dict, cls=NumpyJSONEncoder)

        return ordered_dict

"""
from column_info import ColumnInfo
for item in ColumnInfo.get_variable_labels():
    print (item[0])
"""



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tworavens_preprocess",
    version="0.1.3",
    author="TwoRavens Developers",
    author_email="raman_prasad@harvard.edu",
    description="TwoRavens Preprocess package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TwoRavens/raven-metadata-service",

    packages=['raven_preprocess',
              'raven_preprocess.basic_utils'
              ],

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        'Programming Language :: Python :: 3.6'
        ],

        keywords='tworavens preprocess metadata',  # Optional

        python_requires='>=3.6',

        install_requires=[
            'pandas>=0.22.0',
            'scipy>=1.0.0',
            'simplejson>=3.13.2',
            'xlrd>=1.1.0',
            'jsonschema>=2.6.0',
        ],
)



import numpy as np
import pandas as pd

from raven_preprocess.column_info import ColumnInfo
import raven_preprocess.col_info_constants as col_const


class SummaryStatsUtil(object):

    """This module does the calculation of statistics variables."""
    def __init__(self, col_series, col_info):
        assert isinstance(col_series, pd.Series), "col_series must be a pandas.Series object"
        assert isinstance(col_info, ColumnInfo), "col_info must be a ColumnInfo object"

        self.col_info = col_info
        self.colname = self.col_info.colname
        self.col_series = col_series
        self.total_agg = 0
        self.calc_stats()

    def calc_stats(self):
        """This method does the calculation of statistics variables."""
        self.col_series.dropna(inplace=True)
        # --------------------------
        # similar to preprocess.R "Mode" function
        # --------------------------
        self.col_info.uniques = len(self.col_series.unique())

        mid_pt = int(self.col_info.uniques / 2)

        # iterate through value_counts for mode stats
        #
        row_num = 0
        cnt_min = None
        val_min = None
        cnt_max = None
        val_max = None
        output = []
        fewest_output = []

        for col_val, val_cnt in self.col_series.value_counts(sort=True, ascending=True).iteritems():
            if cnt_min is None and val_min is None and cnt_max is None and val_max is None:
                cnt_min = val_cnt
                val_min = col_val
                cnt_max = val_cnt
                val_max = col_val
            self.total_agg = self.total_agg + val_cnt
            row_num += 1
            if row_num == mid_pt:

                self.col_info.mid = col_val
                self.col_info.freqmid = val_cnt

            if val_cnt == cnt_max:
                output.append(col_val)
                self.col_info.freqmode = val_cnt
            elif val_cnt > cnt_max:
                cnt_max = val_cnt
                output.clear()
                output.append(col_val)
                self.col_info.freqmode = val_cnt

            if val_cnt == cnt_min:
                fewest_output.append(col_val)
                self.col_info.freqfewest = val_cnt
            elif val_cnt < cnt_min:
                cnt_min = val_cnt
                fewest_output.clear()
                fewest_output.append(col_val)
                self.col_info.freqfewest = val_cnt

        self.col_info.set_fewest(fewest_output)

        self.col_info.set_mode(output)

        if self.col_info.is_character():

            # self.col_info.herfindahl=self.herfindahl_index(self.col_series)
            self.col_info.median = col_const.NOT_APPLICABLE
            self.col_info.max = col_const.NOT_APPLICABLE
            self.col_info.min = col_const.NOT_APPLICABLE
            self.col_info.mean = col_const.NOT_APPLICABLE
            self.col_info.std_dev = col_const.NOT_APPLICABLE
            self.col_info.herfindahl = self.herfindahl_index(
                self.col_series,
                True,
                self.total_agg,
                drop_missing=False)

        elif self.col_info.is_numeric():

            self.col_info.median = self.col_series.median()
            self.col_info.max = self.col_series.max()
            self.col_info.min = self.col_series.min()
            self.col_info.mean = self.col_series.mean()
            self.col_info.std_dev = self.col_series.std()
            self.col_info.herfindahl = self.herfindahl_index(
                                                self.col_series,
                                                False,
                                                self.total_agg,
                                                drop_missing=False)

            # self.col_info.mode = np.around(self.col_info.mode, 4)
            if self.col_info.mid is not None:
                self.col_info.mid = np.around(self.col_info.mid, 4)
            # freqfewest and freqmid left for now as they always give int value. why SignIf then?
            # print("--"*20)
            # print("name : ", self.col_info.colname)
            # print("mode : ", self.col_info.mode)
            # print("fewest : ", self.col_info.fewest)
            # print("mid : ", self.col_info.mid)

    @staticmethod
    def herfindahl_index(col_data, char, sum_val,  drop_missing=True):
        # check again with the logic of calculating, what values are squared
        """Calculate Herfindahl-Hirschman Index (HHI) for the column data.
        For each given day, HHI is defined as a sum of squared weights of
        %values in a col_series; and varies from 1/N to 1.
        """
        fraction_val = []
        total_sum = 0
        if drop_missing:
            # redundant if not used as a staticmethod,
            # already happens at calc_stats init
            col_data.dropna(inplace=True)
        if char:
            total_sum = sum_val
            if total_sum == 0:
                pass
            else:
                for val, val_cnt in col_data.value_counts().iteritems():
                    fraction_val.append(\
                        np.math.pow(val_cnt / total_sum, 2))
        else:
            total_sum = sum(col_data)
            if total_sum == 0:
                pass
            else:
                for val, cnt in col_data.items():
                    fraction_val.append(\
                        np.math.pow(cnt / total_sum, 2))

        return sum(fraction_val)



"""Convenience methods for returning namedtuple objects from functions

Error:
- Old "Error" tuple:
    `return (False, 'some error message')`
- New "Error" tuple
    `return err_resp('some error message')`

Success:
- Old "Success" tuple:
    `return (True, some_obj)`
- New "Success" tuple
    `return ok_resp(some_obj)`


"""
from collections import namedtuple


SuccessResponse = namedtuple('SuccessResponse', 'success result_obj')
ErrorResponse = namedtuple('ErrorResponse', 'success err_msg')
ErrorResponseWithData = namedtuple('ErrorResponseWithData', 'success err_msg err_data')

def ok_resp(result_obj):
    """Return a SuccessResponse with success=True and result_obj"""
    return SuccessResponse(True, result_obj)

def err_resp(err_msg):
    """Return a ErrorResponse with success=False and message"""
    return ErrorResponse(False, err_msg)

def err_resp_with_data(err_msg, err_data):
    """Return a ErrorResponse with success=False, a message, and data"""
    return ErrorResponseWithData(False, err_msg, err_data)



"""Constants used to identify file extensions and get mime types"""

CSV_FILE_EXT = '.csv'
TAB_FILE_EXT = '.tab'
XLS_FILE_EXT = '.xls'
XLSX_FILE_EXT = '.xlsx'

ACCEPTABLE_FILE_TYPE_EXTS = (CSV_FILE_EXT,
                             TAB_FILE_EXT,
                             XLS_FILE_EXT,
                             XLSX_FILE_EXT)

ACCEPTABLE_EXT_LIST = ', '.join(['"%s"' % x for x in ACCEPTABLE_FILE_TYPE_EXTS])

MIME_TYPE_PAIRS = (\
    (CSV_FILE_EXT, 'text/csv'),
    (TAB_FILE_EXT, 'text/tab-separated-values'),
    (XLS_FILE_EXT, 'application/vnd.ms-excel'),
    (XLSX_FILE_EXT, ('application/vnd.openxmlformats-officedocument'
                     '.spreadsheetml.sheet')),)

MIME_TYPE_LOOKUP = {x:y for x, y in MIME_TYPE_PAIRS}

def get_mime_type(file_ext):
    """Return a mimetype based on an extension"""
    return MIME_TYPE_LOOKUP.get(\
                file_ext,
                "Unknown mime type for extension: %s" % file_ext)



"""Update the version number within a preprocess dict"""
from decimal import Decimal
import raven_preprocess.col_info_constants as col_const


class VersionNumberUtil(object):
    """Assist with version numbers"""

    @staticmethod
    def update_version_number(preprocess_json, is_major_version=False):
        """Increment the version number by .1
        If "is_major_update", go to the next whole number, e.g. 1.1 -> 2.0
        """
        assert isinstance(preprocess_json, dict), \
            "preprocess_json must be a dict or OrderedDict"

        try:
            version = Decimal(\
                preprocess_json[col_const.SELF_SECTION_KEY][col_const.VERSION_KEY])
        except KeyError as err_obj:
            return False, "version not found: %s" % err_obj

        if is_major_version:
            preprocess_json[col_const.SELF_SECTION_KEY][col_const.VERSION_KEY] = \
                Decimal(int(version + Decimal('1.0')))
        else:
            version += Decimal('.1')
            preprocess_json[col_const.SELF_SECTION_KEY][col_const.VERSION_KEY] = \
                version

        return True, preprocess_json
